package jahuwaldt.jatex;

import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.swing.*;
import java.awt.Frame;
import java.awt.desktop.*;
import java.awt.event.ActionEvent;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import static java.util.Objects.isNull;
import java.util.ResourceBundle;
import javax.swing.*;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

/**
 * This class represents a multi-document interface GUI application.
 * 
 * <p> Modified by: Joseph A. Huwaldt   </p>
 * 
 * @author Joseph A. Huwaldt, Date: May 10, 2013
 * @version December 16, 2023
 */
public class AppGUI extends MDIApplication {

    /**
     * Constructor for our application that displays the GUI.
     * 
     * @param resBundle The resource bundle for this application.
     * @throws java.lang.Exception if anything can not be set up correctly.
     */
    public AppGUI(ResourceBundle resBundle) throws Exception {
        super(resBundle.getString("appName"));
        
        // Set the system look and feel to hide that hideous Java LAF.
        AppUtilities.setSystemLAF();

        // MacOS X users like the menu bar to be at the top of the screen.
        // This does nothing on non-MacOS platforms.
        System.setProperty("apple.laf.useScreenMenuBar", "true");
        
        // Store the resource bundle.
        setResourceBundle(resBundle);

        // Set up the file name filter.
        ExtFilenameFilter fnFilter = new ExtFilenameFilter();

        //  Add any other extensions that might be encountered that we might be able to read.
        String fileExtensions = resBundle.getString("fileExtensions");
        String[] tokens = fileExtensions.split(",");
        for (String token : tokens) {
            fnFilter.addExtension(token.trim());
        }
        setFilenameFilter(fnFilter);

        //  Create the menu bar.
        JMenuBar menuBar = createMenuBar();

        // Add the menu bar to this application.
        //  Cause the application to stay open when the last window is closed under MacOS.
        //  When the last window is closed, a menu bar will remain for the application with no
        //  window open.  This is standard MacOS behavior and will do nothing on other platforms.
        setFramelessJMenuBar(menuBar);
        
    }

    //-----------------------------------------------------------------------------------
    /**
     * Handle the user choosing the "New..." from the File menu. Creates and
     * returns a reference to a new document window.
     * 
     * @param event The event that caused this method to be called.
     * @return A new instance of this application's main window.
     */
    @Override
    public Frame handleNew(ActionEvent event) {
        Frame window = null;

        try {
            // Create an instance of an application window to get the program rolling.
            ResourceBundle resBundle = getResourceBundle();
            String title = resBundle.getString("appName") + ", "
                + resBundle.getString("version") + " "
                + resBundle.getString("appVersion");
            window = new MainWindow(title);
            window.setVisible(true);
            addWindow(window);

        } catch (Exception e) {
            AppUtilities.showException(null, getResourceBundle().getString("unexpectedTitle"),
                    getResourceBundle().getString("unexpectedMsg"), e);
            e.printStackTrace();
        }

        return window;
    }

    /**
     * Create an about handler for use in this application.
     * 
     * @return The about menu item for this program.
     */
    public AboutHandler createAboutHandler() {

        AboutHandler about = new AboutHandler() {
            @Override
            public void handleAbout​(AboutEvent e) {
                ResourceBundle resBundle = getResourceBundle();

                //  Load in the application's icon image.
                Icon appIcon = getApplicationIcon();
                if (isNull(appIcon)) {
                    //  Use a generic icon since this app can't read it's custom icon.
                    appIcon = UIManager.getIcon("OptionPane.informationIcon");
                }

                //  Read in the about box text information.
                String credits = readAboutText(resBundle.getString("aboutTextURLStr"));

                String aName = resBundle.getString("appName");
                String aVersion = resBundle.getString("appVersion");
                String aModDate = resBundle.getString("appModDate");
                String copyright = resBundle.getString("copyright");
                StandardMacAboutFrame f = new StandardMacAboutFrame(aName, aVersion + " - " + aModDate);
                f.setApplicationIcon(appIcon);
                f.setCopyright(copyright);
                f.setCredits(credits, "text/html");
                f.setHyperlinkListener(new HyperlinkListener() {
                    @Override
                    public void hyperlinkUpdate(HyperlinkEvent e) {
                        if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                            //  Display the selected URL in the default browser for this platform.
                            try {
                                MDIApplication.browse(e.getURL().toURI());
                            } catch (Exception err) {
                                err.printStackTrace();
                            }
                        }
                    }
                });
                f.setVisible(true);
            }
        };

        return about;
    }

    /**
     * Return the application icon for this program or null if one can not be found.
     * 
     * @return The application icon for this program.
     */
    public Icon getApplicationIcon() {
        ResourceBundle resBundle = getResourceBundle();
        
        //  Load in the application's icon image.
        Icon appIcon = null;
        try {
            URL imgURL = ClassLoader.getSystemResource(resBundle.getString("applicationIconURL"));
            appIcon = new ImageIcon(imgURL);
        } catch (Exception ex) {
            //LOGGER.log(Level.WARNING, null, ex);
            ex.printStackTrace();
        }
        
        return appIcon;
    }

    /**
     * Method that handles reading in the contents of the text region of the
     * about box from a text file.
     */
    private String readAboutText(String urlStr) {
        String text;
        try {

            URL url = ClassLoader.getSystemResource(urlStr);

            InputStreamReader reader = new InputStreamReader(url.openStream(), "UTF-8");
            StringWriter writer = new StringWriter();
            int character = reader.read();
            while (character != -1) {
                writer.write(character);
                character = reader.read();
            }
            text = writer.toString();

        } catch (IOException e) {
            e.printStackTrace();

            //  Create a standard set of credits if we didn't find the text file.
            text = "<html><body><b>Author:</b><br>Joseph A. Huwaldt<br/>";
            text += "<a href=\"mailto:jhuwaldt@mac.com\">jhuwaldt@mac.com</a><br/>";
            text += "<P ALIGN=CENTER><br/>" + getResourceBundle().getString("appName") + " comes with ABSOLUTELY NO WARRANTY;";
            text += "</body></html>";
        }

        return text;
    }

    /**
     * Return a preferences handler that displays the preference dialog for this application.
     * 
     * @return A preferences handler
     */
    public PreferencesHandler createPreferencesHandler() {
        PreferencesHandler handler = new PreferencesHandler() {
            @Override
            public void handlePreferences(PreferencesEvent e) {
                MDIApplication.getPreferences().showPreferenceDialog();
            }
        };
        return handler;
    }
    
    /**
     * Initializes and displays the menus associated with this application.
     */
    private JMenuBar createMenuBar() throws NoSuchMethodException {
        ResourceBundle resBundle = getResourceBundle();

        //  Create a menu bar.
        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        List<String[]> menuStrings = new ArrayList();
        String[] row = new String[3];
        row[0] = resBundle.getString("newItemText");
        row[1] = resBundle.getString("newItemKey");
        row[2] = "handleNew";
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("openItemText");
        row[1] = resBundle.getString("openItemKey");
        row[2] = null;
        menuStrings.add(row);
        menuStrings.add(new String[3]);  //  Blank line
        row = new String[3];
        row[0] = resBundle.getString("closeItemText");
        row[1] = resBundle.getString("closeItemKey");
        row[2] = "handleClose";
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("saveItemText");
        row[1] = resBundle.getString("saveItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("saveAsItemText");
        row[1] = null;
        row[2] = null;
        menuStrings.add(row);

        JMenu menu = AppUtilities.buildMenu(this, resBundle.getString("fileMenuText"), menuStrings);
        menuBar.add(menu);

        //  Add a Quit menu item (if it isn't automatically present).
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(this.getQuitJMenuItem());
        }
        
        //  Add a Quit handler to ask the user if they want to save unsaved changes before quitting.
        setQuitHandler(new QuitHandler() {
            @Override
            public void handleQuitRequestWith​(QuitEvent e, QuitResponse response) {
                handleQuit(e, response);
            }
        });
        
        
        // Set up the edit menu.
        menuStrings.clear();
        row = new String[3];
        row[0] = resBundle.getString("undoItemText");
        row[1] = resBundle.getString("undoItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("redoItemText");
        row[1] = null;
        row[2] = null;
        menuStrings.add(row);
        menuStrings.add(new String[3]);  //  Blank line
        row = new String[3];
        row[0] = resBundle.getString("cutItemText");
        row[1] = resBundle.getString("cutItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("copyItemText");
        row[1] = resBundle.getString("copyItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("pasteItemText");
        row[1] = resBundle.getString("pasteItemKey");
        row[2] = null;
        menuStrings.add(row);

        menu = AppUtilities.buildMenu(this, resBundle.getString("editMenuText"), menuStrings);
        menuBar.add(menu);

        //  Register a Preferences menu item handler.
        setPreferencesHandler(createPreferencesHandler());
        
        //  Create a Preferences menu item (if not automatically present).
        if (!PreferencesJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(this.getPreferencesJMenuItem());
        }

        //	Register an about menu item handler.
        setAboutHandler(createAboutHandler());
        
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(resBundle.getString("helpMenuText"));
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(this.getAboutJMenuItem());
        }
        
        return menuBar;
    }

}
