/**
 * Please feel free to use any fragment of the code in this file that you need in your own
 * work. As far as I am concerned, it's in the public domain. No permission is necessary
 * or required. Credit is always appreciated if you use a large chunk or base a
 * significant product on one of my examples, but that's not required either.
 * 
 * This code is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.
 * 
 * --- Joseph A. Huwaldt
 */
package jahuwaldt.swing;

import java.io.*;
import java.util.Locale;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import javax.swing.JFrame;
import javax.swing.JRootPane;



/**
 * A set of utilities that are used by my programs when running under MacOS only. These
 * static methods provide Mac specific functionality but all the methods in it can be
 * called safely from any platform. On non-Mac systems, the worst that will happen when
 * calling these methods is nothing and sometimes appropriate non-Mac behavior is
 * provided.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2000
 * @version December 3, 2023
 */
public class MacOSUtilities {

	private static final boolean IS_MACOS;

    static {
        String OS = System.getProperty("os.name", "generic").toLowerCase(Locale.ENGLISH);
        IS_MACOS = (OS.contains("mac") || OS.contains("darwin"));
    }
    
	/**
	*  Returns true if this program is running in any MacOS 8/9/X environment,
	*  false is returned otherwise.
	**/
	public static boolean isMacOS() {
		return IS_MACOS;
	}

	/**
	*  Returns the a File reference to the specified resource in the MacOS X application
	*  bundle's Resource directory.  This method takes advantage of the application
	*  bundle's used by MacOS X to hide resources from the user.
	*  If not running on a MacOS X system, then the resource is searched for in three
	*  locations.  1st the directory that the application is running in.  2nd the
	*  default application name supplied is combined with
	*  ".app/Contents/Resources/" + resourceName and searched for in the current directory.
	*  This is how a MacOS X application bundle would appear on a non-Mac system.
	*  Third, the default application name is combined with ".app/Contents/Resources/" + resourceName
	*  and searched for in the install directory (if it can be determined).
	*
	*  @param  resourceName   The name of the resource file.
	*  @param  defaultAppName The default application name to use if not running on MacOS.
	*  @throws FileNotFoundException if the specified resource file could not be found.
	**/
	public static File getResource(String resourceName, String defaultAppName) throws FileNotFoundException {

        //	First look for the resource in the application directory.
        File file = new File(resourceName);

        if (!file.exists()) {

            //	Try looking in a MacOS X style application bundle in current (working) directory.
            String sep = System.getProperty("file.separator");
            StringBuilder path = new StringBuilder(defaultAppName);
            path.append(".app");
            path.append(sep);
            path.append("Contents");
            path.append(sep);
            path.append("Resources");
            path.append(sep);
            path.append(resourceName);
            file = new File(path.toString());

            //	Try looking for MacOS X style application bundle in install directory (if known).
            if (!file.exists()) {
                String installPath = System.getProperty("lax.root.install.dir");
                if (installPath == null || installPath.equals(""))
                    installPath = System.getProperty("user.dir");

                installPath = installPath.replace(File.separatorChar, '/');
                if (!installPath.endsWith("/"))
                    path.insert(0, "/");
                path.insert(0, installPath);
                file = new File(path.toString());

                //	Couldn't find the resource file anywhere.
                if (!file.exists())
                    throw new FileNotFoundException(path.toString());
            }

        }
		
		return file;
	}
	
	/**
	*  Returns the a File reference to the specified resource in the specified
	*  sub-directory of the application bundle's Resource directory.  This
	*  method takes advantage of the application bundle's used by MacOS X to
	*  hide resources from the user.
	*  If not running on a MacOS X system, then the resource is searched for in three
	*  locations.  1st the directory that the application is running in.  2nd the
	*  default application name supplied is combined with
	*  ".app/Contents/Resources/" + resourceName and searched for in the current directory.
	*  This is how a MacOS X application bundle would appear on a non-Mac system.
	*  Third, the default application name is combined with ".app/Contents/Resources/" + resourceName
	*  and searched for in the install directory (if it can be determined).
	*
	*  @param  resourceName   The name of the resource file.
	*  @param  subDirName     The name of the sub-directory under Resources.
	*  @param  defaultAppName The default application name to use if not running on MacOS.
	*  @throws FileNotFoundException if the specified resource file could not be found.
	**/
	public static File getResource(String resourceName, String subDirName, String defaultAppName)
															throws FileNotFoundException {

        //	First look for the resource in the application directory.
        String sep = System.getProperty("file.separator");
        StringBuffer path = new StringBuffer(subDirName);
        path.append(sep);
        path.append(resourceName);
        File file = new File(path.toString());

        if (!file.exists()) {

            //	Try looking in a MacOS X style application bundle in current (working) directory.
            path = new StringBuffer(defaultAppName);
            path.append(".app");
            path.append(sep);
            path.append("Contents");
            path.append(sep);
            path.append("Resources");
            path.append(sep);
            path.append(subDirName);
            path.append(sep);
            path.append(resourceName);
            file = new File(path.toString());

            //	Try looking for MacOS X style application bundle in install directory (if known).
            if (!file.exists()) {
                String installPath = System.getProperty("lax.root.install.dir");
                if (installPath == null || installPath.equals(""))
                    installPath = System.getProperty("user.dir");

                installPath = installPath.replace(File.separatorChar, '/');
                if (!installPath.endsWith("/"))
                    path.insert(0, "/");
                path.insert(0, installPath);
                file = new File(path.toString());

                //	Couldn't find the resource file anywhere.
                if (!file.exists())
                    throw new FileNotFoundException(path.toString());
            }
        }
		
		return file;
	}
	
	
    /**
     * Sets the "Window.documentModified" client property on the frame's root pane to the
     * value indicated. On a Mac, the modified-document mark is standardized as a black
     * dot in the center of the red close button on the title bar. The proxy icon, if
     * present, is also dimmed if <code>modified</code> is set to true.
     *
     * @param window The JFrame to have the modified mark set on.
     * @param modified Set to true to indicate that the document represented by frame is
     * modified. Set to false to indicate that it has not been modified since it was last
     * created, read in, or saved.
     */
    public static void setModifiedMark(JFrame window, boolean modified) {
        JRootPane root = window.getRootPane();
        root.putClientProperty("Window.documentModified", modified);
    }
    
    /**
     * Return the state of the "Window.documentModified" client property for the specified
     * frame.  On a Mac, the modified-document mark is standardized as a black
     * dot in the center of the red close button on the title bar.  It is used to indicate
     * that a document has been modified.
     * 
     * @param window  The frame to have the document modified client property read from.
     */
    public static Boolean isMarkedModified(JFrame window) {
        JRootPane root = window.getRootPane();
        Object prop = root.getClientProperty("Window.documentModified");
        if (nonNull(prop) && prop instanceof Boolean) {
            return (Boolean)prop;
        }
        return Boolean.FALSE;
    }

    /**
     * Set's the title bar proxy icon. On a Mac, document windows have a proxy icon to the
     * left of the title text. Users can click and drag the icon to other windows or
     * right-click/command-click to get a path to the document.
     *
     * @param window The JFrame to have a proxy icon added to the title bar.
     * @param file The file who's icon should be added to the title bar. The Mac
     * automatically finds the correct icon for the file or folder.
     */
    public static void setProxyIcon(JFrame window, File file) {
        JRootPane root = window.getRootPane();
        root.putClientProperty("Window.documentFile", file);
    }
    
    /**
     * Return the contents of the "Window.documentFile" client property (which can
     * return null).  On a Mac, document windows have a proxy icon to the
     * left of the title text. Users can click and drag the icon to other windows or
     * right-click/command-click to get a path to the document.  The file who's icon
     * is used as the proxy icon, is retrieved from a frame with this method.
     * 
     * @param window The JFrame to have the proxy icon file retrieved.
     * @return The file who's icon is used as the proxy icon or null if there is none.
     */
    public static File getProxyIcon(JFrame window) {
        JRootPane root = window.getRootPane();
        return (File)root.getClientProperty("Window.documentFile");
    }
}
